/*
 * Copyright (c) 2008-2009, Maconnect
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code MUST retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Maconnect nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#import "FVPUID.h"
#import <string>
#import "ofa.h"

using namespace std;

// This object must be filled out completely prior to making any
// calls to the server.  On return, some fields will be filled out.
class TrackInformation {
private:
    string puid;
    string encoding;        // All other strings must honor this encoding
    int    bitrate;         // i.e. "192kbps", use 0 for VBR or freeformat
    string format;          // File extension
    long   length_in_ms;    // In milliseconds
    string artist;
    string track;
    string album;
    int    trackNum;        // use 0 if not known
    string genre;
    string year;
public:
    string print;
    TrackInformation() :
	bitrate(0), length_in_ms(0), trackNum(0) {}
    ~TrackInformation() {}
    void setPrint(string p) { print = p; }
    string getPrint() const { return print; }
    // Only supported encodings are UTF-8 (default) and ISO-8859-15
    void setEncoding(string e) { encoding = e; }
    string getEncoding() const { return encoding; }
    void setBitrate(int b) { bitrate = b; }
    int getBitrate() const { return bitrate; }
    void setFormat(string fmt) { format = fmt; }
    string getFormat() const { return format; }
    void setLengthInMS(long ms) { length_in_ms = ms; }
    long getLengthInMS() const { return length_in_ms; }
    void setArtist(string name) { artist = name; }
    string getArtist() const { return artist; }
    void setTrack(string name) { track = name; }
    string getTrack() const { return track; }
    void setAlbum(string name) { album = name; }
    string getAlbum() const { return album; }
    void setTrackNum(int t) { trackNum = t; }
    int getTrackNum() const { return trackNum; }
    void setGenre(string g) { genre = g; }
    string getGenre() const { return genre; }
    void setYear(string y) { year = y; }
    string getYear() const { return year; }
    void setPUID(string id)  { puid = id; }
    string getPUID() const { return puid; }
};

// Get your unique key at http://www.musicdns.org
bool retrieve_metadata(string client_key, string client_verstion,
					   TrackInformation *info, bool getMetadata);

class AudioData {
private:
    unsigned char *samples;
    int byteOrder;
    long size;
    int sRate;
    bool stereo;
public:
    TrackInformation info;
    AudioData() : samples(0), size(0), sRate(0), stereo(false) {}
    ~AudioData() {
		delete[] samples;
    }
    // size is number of samples (half the number of bytes)
    void setData(unsigned char*_samples, int _byteOrder, long _size,
				 int _sRate, bool _stereo, int _ms, string _fmt) {
		samples = _samples;
		byteOrder = _byteOrder;
		size = _size;
		sRate = _sRate;
		stereo = _stereo;
		// These two fields are used later for the protocol layer
		info.setLengthInMS(_ms);
		info.setFormat(_fmt);
    }
    int getByteOrder() const { return byteOrder; }
    long getSize() const { return size; }
    int getSRate() const { return sRate; }
    bool getStereo() const { return stereo; }
    bool createPrint() {
		const char *print = ofa_create_print(samples, byteOrder, size, sRate, stereo);
		if (!print)
			return false;
		info.setPrint(print);
		return true;
    }
    // Get your unique key at http://www.musicdns.org
    TrackInformation *getMetadata(string client_key, string client_version, 
								  bool metadataFlag)
    {
		if (!retrieve_metadata(client_key, client_version, &info, metadataFlag))
			return 0;
		return &info;
    }
};


const char *unknown = "unknown";
// Lookup by fingerprint
const char *request_format = 
"cid=%@&"       // Client ID
"cvr=%@&"       // Client Version
"fpt=%s&"       // Fingerprint
"rmd=%d&"       // m = 1: return metadata; m = 0: only return id 
"brt=%d&"       // bitrate (kbps)
"fmt=%s&"       // File extension (e.g. mp3, ogg, flac)
"dur=%ld&"      // Length of track (milliseconds)
"art=%s&"       // Artist name. If there is none, send "unknown"
"ttl=%s&"       // Track title. If there is none, send "unknown"
"alb=%s&"       // Album name. If there is none, send "unknown"
"tnm=%d&"       // Track number in album. If there is none, send "0"
"gnr=%s&"       // Genre. If there is none, send "unknown"
"yrr=%s&"       // Year. If there is none, send "0"
//@"enc=%s&"       // Encoding. e = true: ISO-8859-15; e = false: UTF-8 (default). Optional.
"\r\n";

static bool readBytes(int fd, unsigned char *buf, int size) {
    int ct = 0;
    while (ct < size) {
		unsigned char tmp[4096];
		
		int x = size - ct;
		if (x > 4096)
			x = 4096;
		
		int n = read(fd, tmp, x);
		
		if (n <= 0) {
			return false;
		}
		
		for (int i = 0; i < n; ++i) {
			buf[ct + i] = tmp[i];
		}
		ct += n;
    }
    return true;
}

// This method only supports PCM/uncompressed format, with a single fmt
// chunk followed by a single data chunk
AudioData* loadWaveFile(const char *file) {
	
    int srate = 0;
    int channels = 0;
	
    int fd = open(file, O_RDONLY | 0x8000);
    if (fd == -1)
		return 0;
	
    if (lseek(fd, 0L, SEEK_SET) == -1L) {
		close(fd);
		return 0;
    }
	
    unsigned char hdr[36];
    if (!readBytes(fd, hdr, 36)) {
		close(fd);
		return 0;
    }
    if (hdr[0] != 'R' || hdr[1] != 'I' || hdr[2] != 'F' || hdr[3] != 'F') {
		close(fd);
		return 0;
    }
    // Note: bytes 4 thru 7 contain the file size - 8 bytes
    if (hdr[8] != 'W' || hdr[9] != 'A' || hdr[10] != 'V' || hdr[11] != 'E') {
		close(fd);
		return 0;
    }
    if (hdr[12] != 'f' || hdr[13] != 'm' || hdr[14] != 't' || hdr[15] != ' ') {
		close(fd);
		return 0;
    }
	
    long extraBytes = hdr[16] + (hdr[17] << 8) + (hdr[18] << 16) + (hdr[19] << 24) - 16;
    int compression = hdr[20] + (hdr[21] << 8);
    // Type 1 is PCM/Uncompressed
    if (compression != 1) {
		close(fd);
		return 0;
    }
    channels = hdr[22] + (hdr[23] << 8);
    // Only mono or stereo PCM is supported in this example
    if (channels < 1 || channels > 2) {
		close(fd);
		return 0;
    }
    // Samples per second, independent of number of channels
    srate = hdr[24] + (hdr[25] << 8) + (hdr[26] << 16) + (hdr[27] << 24);
    // Bytes 28-31 contain the "average bytes per second", unneeded here
    // Bytes 32-33 contain the number of bytes per sample (includes channels)
    // Bytes 34-35 contain the number of bits per single sample
    int bits = hdr[34] + (hdr[35] << 8);
    // Supporting othe sample depths will require conversion
    if (bits != 16) {
		close(fd);
		return 0;
    }
	
    // Skip past extra bytes, if any
    if (lseek(fd, 36L + extraBytes, SEEK_SET) == -1L) {
		close(fd);
		return 0;
    }
	
    // Start reading the next frame.  Only supported frame is the data block
    unsigned char b[8];
    if (!readBytes(fd, b, 8)) {
		close(fd);
		return 0;
    }
    // Do we have a fact block?
    if (b[0] == 'f' && b[1] == 'a' && b[2] == 'c' && b[3] == 't') {
		// Skip the fact block
		if (lseek(fd, 36L + extraBytes + 12L, SEEK_SET) == -1L) {
			close(fd);
			return 0;
		}
		// Read the next frame
		if (!readBytes(fd, b, 8)) {
			close(fd);
			return 0;
		}
    }
	
    // Now look for the data block
    if (b[0] != 'd' || b[1] != 'a' || b[2] != 't' || b[3] != 'a') {
		close(fd);
		return 0;
    }
    long bytes = b[4] + (b[5] << 8) + (b[6] << 16) + (b[7] << 24);
	
    long ms = (bytes/2)/(srate/1000);
    if ( channels == 2 ) ms /= 2;
	
    // No need to read the whole file, just the first 135 seconds
    int sampleSize = 135;
    long bytesInNSecs = sampleSize * srate * 2 * channels;
    bytes = bytes > bytesInNSecs ? bytesInNSecs : bytes;
	
    unsigned char *samples = new unsigned char[bytes];
    if (!readBytes(fd, samples, bytes)) {
		delete[] samples;
		close(fd);
		return 0;
    }
    close(fd);
	
    AudioData *data = new AudioData();
    data->setData(samples, OFA_LITTLE_ENDIAN, bytes/2, srate, channels == 2 ? 1 : 0, ms, "wav");
    return data;
}

@implementation FVPUID

+(NSArray*)listOfPUIDsForWavFileAtPath:(NSString*)path duration:(float)duration clientID:(NSString*)cid version:(NSString*)version
{
	NSMutableArray	*listOfPUIDs=NULL;
	AudioData		*data=loadWaveFile([path fileSystemRepresentation]);
	if (data->createPrint()){
		NSMutableURLRequest	*request=[NSMutableURLRequest
									  requestWithURL:[NSURL URLWithString:@"http://ofa.musicdns.org/ofa/1/track"]];
		
		[request setHTTPMethod:@"POST"];
        NSString            *strFormat=[NSString stringWithUTF8String:request_format];
		NSString			*body=[NSString stringWithFormat:strFormat,
								   cid,
								   version,
								   data->info.print.c_str(),
								   0,
								   data->info.getBitrate(),
								   data->info.getFormat().c_str(), 
								   duration > 0.1 ? (long)(duration*1000.) : data->info.getLengthInMS(), 
								   (data->info.getArtist().length() == 0) ? unknown : data->info.getArtist().c_str(),
								   (data->info.getTrack().length() == 0) ? unknown : data->info.getTrack().c_str(),
								   (data->info.getAlbum().length() == 0) ? unknown : data->info.getAlbum().c_str(),
								   data->info.getTrackNum(), 
								   (data->info.getGenre().length() == 0) ? unknown : data->info.getGenre().c_str(),
								   (data->info.getYear().length() == 0) ? "0" : data->info.getYear().c_str(),
								   @"false"
								   ];
		[request setHTTPBody:[body dataUsingEncoding:NSUTF8StringEncoding]];
		
		NSURLResponse	*resp=NULL;
		NSData			*data=[NSURLConnection sendSynchronousRequest:request returningResponse:&resp error:NULL];
		if (data){
			NSXMLDocument	*xml=[[NSXMLDocument alloc] initWithData:data options:0 error:NULL];
			if (xml)
			{
				[xml autorelease];
				
				NSXMLElement	*root=[xml rootElement];
				NSArray			*a=[root elementsForName:@"track"];
				if (a && [a count])
				{
					a=[[a objectAtIndex:0] elementsForName:@"puid-list"];
					if (a && [a count])
					{
						a=[[a objectAtIndex:0] elementsForName:@"puid"];
						if (a && [a count])
						{
							listOfPUIDs=[NSMutableArray arrayWithCapacity:1];
							NSEnumerator	*e=[a objectEnumerator];
							NSXMLElement	*elem=NULL;
							
							while((elem=[e nextObject]))
							{
								NSXMLNode	*node=[elem attributeForName:@"id"];
								if (!node)
									continue;
								id			value=[node objectValue];
								if ([value isKindOfClass:[NSString class]])
									[listOfPUIDs addObject:value];
							}
						}
					}
				}
			}
		}
	}
	
	return listOfPUIDs;
}

@end
