/*
 * Copyright (c) 2010-2011, Maconnect
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code MUST retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Maconnect, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#import <Cocoa/Cocoa.h>
#import "mirageConst.h"
#import "mirage.h"
#include <unistd.h>

static const char				appVersion[] = "1.0.0";
static char						programName[PATH_MAX+1]={0};


static GError					*error=NULL;
#define g_check_error(x, y)		{if(!(x)){fprintf(stderr, "%s\n", error->message);\
g_error_free(error); error=NULL;goto y;}}

NSDictionary* getTrackInfo(MIRAGE_Track* track)
{
	gint						len=0, mode=0, layoutStart=0, start=0;
	gchar						*isrc=NULL;
	NSMutableDictionary			*trackInfo=[NSMutableDictionary dictionaryWithCapacity:2];
	
	g_check_error(mirage_track_layout_get_length(track, &len, &error), bail);
	[trackInfo setObject:[NSNumber numberWithInt:len] forKey:trackInfoLength];
	
	g_check_error(mirage_track_get_mode(track, &mode, &error), bail);
	[trackInfo setObject:[NSNumber numberWithInt:mode] forKey:trackInfoMode];
	
	g_check_error(mirage_track_layout_get_start_sector(track, &layoutStart, &error), bail);
	[trackInfo setObject:[NSNumber numberWithInt:layoutStart] forKey:trackInfoLayoutStart];
	
	g_check_error(mirage_track_get_track_start(track, &start, &error), bail);
	[trackInfo setObject:[NSNumber numberWithInt:start] forKey:trackInfoStart];
	
	if (mirage_track_get_isrc(track, (const gchar**)&isrc, NULL)){
		NSData		*data=[NSData dataWithBytesNoCopy:isrc length:12 freeWhenDone:NO];
		NSString	*s=[[[NSString alloc] initWithData:data encoding:[NSString defaultCStringEncoding]] autorelease];
		[trackInfo setObject:s forKey:trackInfoISRC];
	}
	
	MIRAGE_Sector				*sector=NULL;
	gint						sectorLen=0;
	
	if (mirage_track_get_sector(track, 0, 0, (GObject**)&sector, NULL)){
		if (mirage_sector_get_data(sector, NULL, &sectorLen, NULL)){
			[trackInfo setObject:[NSNumber numberWithInt:sectorLen] forKey:trackInfoSectorSize];
		}
		else {
			sectorLen=0;
		}
		g_object_unref(sector);
	}
	if (!sectorLen)
		return NULL;
	
	return trackInfo;
bail:;
	return NULL;
}

NSDictionary* getSessionInfo(MIRAGE_Session* session){
	gint				trackCount=0, i, type=0;
	
	g_check_error(mirage_session_get_number_of_tracks(session, &trackCount, &error), bail);
	
	NSMutableDictionary	*sessionInfo=[NSMutableDictionary dictionaryWithCapacity:1];
	NSMutableArray		*tracks=[[[NSMutableArray alloc] initWithCapacity:trackCount] autorelease];
	
	g_check_error(mirage_session_get_session_type(session, &type, &error), bail);
	[sessionInfo setObject:[NSNumber numberWithInt:type] forKey:sessionInfoType];
	
	MIRAGE_Track		*track;
	for(i=0; trackCount > i; i++){
		g_check_error(mirage_session_get_track_by_index(session, i, (GObject**)&track, &error), bail);
		NSDictionary	*trackInfo=getTrackInfo(track);
		if (!trackInfo)
			return NULL;
		[tracks addObject:trackInfo];
	}
	
	[sessionInfo setObject:tracks forKey:sessionInfoTracks];
	return sessionInfo;
bail:;
	return NULL;
}

NSDictionary* getDiscInfo(MIRAGE_Disc* disc){
	gint				sessionCount=0;
	gint				i;
	gchar				*mcn=NULL;
	NSMutableDictionary	*layoutDescription=NULL;
	
	if (layoutDescription)
		[layoutDescription release];
	layoutDescription=NULL;
	
	g_check_error(mirage_disc_get_number_of_sessions(disc, &sessionCount, &error), bail);
	layoutDescription=[[[NSMutableDictionary alloc] initWithCapacity:2] autorelease];
	
	if (mirage_disc_get_mcn(disc, (const gchar**)&mcn, NULL)){
		NSData			*data=[NSData dataWithBytesNoCopy:mcn length:12 freeWhenDone:NO];
		NSString		*s=[[[NSString alloc] initWithData:data encoding:[NSString defaultCStringEncoding]] autorelease];
		[layoutDescription setObject:s forKey:discInfoMCN];
	}
	
	MIRAGE_Session		*session;
	NSDictionary		*sessionInfo;
	NSMutableArray		*sessions=[[[NSMutableArray alloc] initWithCapacity:sessionCount] autorelease];
	
	for(i=0; sessionCount > i; i++){
		g_check_error(mirage_disc_get_session_by_index(disc, i, (GObject**)&session, &error), bail);
		
		if ((sessionInfo=getSessionInfo(session))){
			[sessions addObject:sessionInfo];
		}
		else {
			return NULL;
		}
	}
	[layoutDescription setObject:sessions forKey:discInfoSessions];
	
	return layoutDescription;
bail:;
	return NULL;
}

void getProgramName(const char *appPath){
	const char		*startPtr=appPath+strlen(appPath);
	while('/'!=*(startPtr--));
	strcpy(programName, startPtr+2);
}

void printUsage(){
	fprintf(stderr, "This is %s v%s\nInvalid argument(s).\n\n", programName, appVersion);
}

int main(int argc, const char *argv[]){
	NSAutoreleasePool			*pool=[[NSAutoreleasePool alloc] init];
	g_type_init();
    int							opt=0;
	const int					maxInputFiles=20;
	gchar						inputFiles[maxInputFiles][PATH_MAX+1];
	int							inputFilesCount=0;
	BOOL						printInfo=NO;
	BOOL						isInit=NO;
	gchar						*inputFilesPtr[maxInputFiles];
	char						outputFile[PATH_MAX+1]={0};
	long						sessionNum=-1, trackNum=-1;
	MIRAGE_Disc					*disc=NULL;
	
	getProgramName(*argv);
	
	if (!argc){
		printUsage();
		goto bail;
	}
	
	bzero(inputFiles, maxInputFiles+(PATH_MAX+1));
	
	while((opt=getopt(argc, (char**)argv, "i:no:s:t:"))!=-1){
		switch(opt){
			case 'i':
				require(inputFilesCount<maxInputFiles-1, bail);
				strlcpy(inputFiles[inputFilesCount++], optarg, PATH_MAX+1);
				break;
			case 'n':
				printInfo=YES;
				break;
			case 'o':
				strlcpy(outputFile, optarg, PATH_MAX+1);
				break;
			case 't':
				trackNum=strtol(optarg, NULL, 10);
				break;
			case 's':
				sessionNum=strtol(optarg, NULL, 10);
				break;
			default:
				printUsage();
				goto bail;
		}
	}
	
	if (!inputFilesCount){
		printUsage();
		goto bail;
	}
	else {
		int						i;
		for(i=0; inputFilesCount>i; i++)
			inputFilesPtr[i]=inputFiles[i];
		inputFilesPtr[inputFilesCount]=NULL;
	}

	//Initialize libmirage
	NSString					*pluginPath=[[[[NSBundle mainBundle] resourcePath]
											  stringByDeletingLastPathComponent]
											 stringByAppendingPathComponent:@"Mirages"];
	setenv("LIQUIDCD_MIRAGE_PLUGIN_DIR", [pluginPath fileSystemRepresentation], 1);
	if (!libmirage_init(nil)){
		fprintf(stderr, "init failed");
		goto bail;
	}
	isInit=YES;
	error=NULL;
	
	//Open the disc image
	disc = (MIRAGE_Disc*)libmirage_create_disc(inputFilesPtr, nil, nil, &error);
	if (error){
		fprintf(stderr, "%s\n", error->message);
		g_error_free(error);
		goto bail;
	}
	
	if (printInfo){
		//Print disc image info and quit.
		NSDictionary			*discInfo=getDiscInfo(disc);
		require(discInfo, bail);
		fprintf(stdout, "%s\n", [[discInfo description] cStringUsingEncoding:NSASCIIStringEncoding]);
	}
	else {
		if (!outputFile || trackNum<0 || sessionNum<0){
			printUsage();
			goto bail;
		}
		int						fd=open(outputFile, O_WRONLY);
		if (fd < 1){
			fprintf(stderr, "Could not open file\n");
			goto bail;
		}
		//We need to find the requested track
		gint					sessionCount=0, trackCount=0;
		
		g_check_error(mirage_disc_get_number_of_sessions(disc, &sessionCount, &error), bail);
		
		if(sessionCount <= sessionNum){
			goto bail;
		}
		MIRAGE_Session			*session=NULL;
		g_check_error(mirage_disc_get_session_by_index(disc, sessionNum, (GObject**)&session, &error), bail);
		if(!mirage_session_get_number_of_tracks(session, &trackCount, &error) || trackNum>=trackCount){
			g_object_unref(session);
			fprintf(stderr, "%s\n", error->message);
			g_error_free(error);
			goto bail;
		}
		error=NULL;
		
		MIRAGE_Track			*track=NULL;
		if (!mirage_session_get_track_by_index(session, trackNum, (GObject**)&track, &error)){
			g_object_unref(session);
			fprintf(stderr, "%s\n", error->message);
			g_error_free(error);
			goto bail;
		}
		error=NULL;
		
		gint					len=0;
		if (!mirage_track_layout_get_length(track, &len, &error)){
			g_object_unref(track);
			g_object_unref(session);
			fprintf(stderr, "%s\n", error->message);
			g_error_free(error);
			goto bail;
		}
		error=NULL;
		gint					trackStart=0;
		if (!mirage_track_get_track_start(track, &trackStart, &error)){
			g_object_unref(track);
			g_object_unref(session);
			fprintf(stderr, "%s\n", error->message);
			g_error_free(error);
			goto bail;
		}
		error=NULL;
		
		//Let's start writing in the output file
		gint					i, sectorLen;
		GObject					*sector=NULL;
		uint64_t				totalWrittenLen=0;
		
		for(i=trackStart; len>i; i++){
			error=NULL;
			if (!mirage_track_get_sector(track, i, 0, &sector, &error)){
				g_object_unref(track);
				g_object_unref(session);
				fprintf(stderr, "%s (address: %d)\n", error->message, (int)i);
				g_error_free(error);
				goto bail;
			}
			
			const guint8		*theData;
			if (!mirage_sector_get_data(MIRAGE_SECTOR(sector), &theData, &sectorLen, &error)){
				g_object_unref(sector);
				g_object_unref(track);
				g_object_unref(session);
				fprintf(stderr, "%s\n", error->message);
				g_error_free(error);
				goto bail;
			}
			
			totalWrittenLen+=write(fd, (const void*)theData, sectorLen);
			
			g_object_unref(sector);
			sector=NULL;
		}
		//fprintf(stderr, "Info: last address is %d\n", i);
		//fprintf(stderr, "Info: total len is %llu\n", totalWrittenLen);
		
		close(fd);
		g_object_unref(track);
		g_object_unref(session);
	}
	
	if (disc)
		g_object_unref(disc);
	if (isInit)
		libmirage_shutdown(NULL);
	[pool drain];
	return EXIT_SUCCESS;
bail:;
	if (disc)
		g_object_unref(disc);
	if (isInit)
		libmirage_shutdown(NULL);
	[pool drain];
	return EXIT_FAILURE;
}
